from __future__ import absolute_import

import os
import re
import time
import traceback

import opentracing as ot
from basictracer import BasicTracer

from .binary_propagator import BinaryPropagator
from .http_propagator import HTTPPropagator
from .text_propagator import TextPropagator
from .recorder import StandardRecorder, InstanaSampler
from .span import InstanaSpan, RegisteredSpan, SpanContext
from .util import generate_id


class InstanaTracer(BasicTracer):
    def __init__(self, scope_manager=None, recorder=None):

        if recorder is None:
            recorder = StandardRecorder()

        super(InstanaTracer, self).__init__(
            recorder, InstanaSampler(), scope_manager)

        self._propagators[ot.Format.HTTP_HEADERS] = HTTPPropagator()
        self._propagators[ot.Format.TEXT_MAP] = TextPropagator()
        self._propagators[ot.Format.BINARY] = BinaryPropagator()

    def handle_fork(self):
        # Nothing to do for the Tracer;  Pass onto Recorder
        self.recorder.handle_fork()

    def start_active_span(self,
                          operation_name,
                          child_of=None,
                          references=None,
                          tags=None,
                          start_time=None,
                          ignore_active_span=False,
                          finish_on_close=True):

        # create a new Span
        span = self.start_span(
            operation_name=operation_name,
            child_of=child_of,
            references=references,
            tags=tags,
            start_time=start_time,
            ignore_active_span=ignore_active_span,
        )

        return self.scope_manager.activate(span, finish_on_close)

    def start_span(self,
                   operation_name=None,
                   child_of=None,
                   references=None,
                   tags=None,
                   start_time=None,
                   ignore_active_span=False):
        "Taken from BasicTracer so we can override generate_id calls to ours"

        start_time = time.time() if start_time is None else start_time

        # See if we have a parent_ctx in `references`
        parent_ctx = None
        if child_of is not None:
            parent_ctx = (
                child_of if isinstance(child_of, SpanContext)
                else child_of.context)
        elif references is not None and len(references) > 0:
            # TODO only the first reference is currently used
            parent_ctx = references[0].referenced_context

        # retrieve the active SpanContext
        if not ignore_active_span and parent_ctx is None:
            scope = self.scope_manager.active
            if scope is not None:
                parent_ctx = scope.span.context

        # Assemble the child ctx
        gid = generate_id()
        ctx = SpanContext(span_id=gid)
        if parent_ctx is not None:
            if parent_ctx._baggage is not None:
                ctx._baggage = parent_ctx._baggage.copy()
            ctx.trace_id = parent_ctx.trace_id
            ctx.sampled = parent_ctx.sampled
        else:
            ctx.trace_id = gid
            ctx.sampled = self.sampler.sampled(ctx.trace_id)

        # Tie it all together
        span = InstanaSpan(self,
                           operation_name=operation_name,
                           context=ctx,
                           parent_id=(None if parent_ctx is None else parent_ctx.span_id),
                           tags=tags,
                           start_time=start_time)

        if operation_name in RegisteredSpan.EXIT_SPANS:
            self.__add_stack(span)

        elif operation_name in RegisteredSpan.ENTRY_SPANS:
            # For entry spans, add only a backtrace fingerprint
            self.__add_stack(span, limit=2)

        return span

    def inject(self, span_context, format, carrier):
        if format in self._propagators:
            return self._propagators[format].inject(span_context, carrier)
        else:
            raise ot.UnsupportedFormatException()

    def extract(self, format, carrier):
        if format in self._propagators:
            return self._propagators[format].extract(carrier)
        else:
            raise ot.UnsupportedFormatException()

    def __add_stack(self, span, limit=None):
        """ Adds a backtrace to this span """
        span.stack = []
        frame_count = 0

        tb = traceback.extract_stack()
        tb.reverse()
        for frame in tb:
            if limit is not None and frame_count >= limit:
                break

            # Exclude Instana frames unless we're in dev mode
            if "INSTANA_DEBUG" not in os.environ:
                if re_tracer_frame.search(frame[0]) is not None:
                    continue

                if re_with_stan_frame.search(frame[2]) is not None:
                    continue

            span.stack.append({
                "c": frame[0],
                "n": frame[1],
                "m": frame[2]
            })

            if limit is not None:
                frame_count += 1


# Used by __add_stack
re_tracer_frame = re.compile(r"/instana/.*\.py$")
re_with_stan_frame = re.compile('with_instana')



""" Options for the in-process Instana agent """
import logging
import os

from .util import determine_service_name

class StandardOptions(object):
    """ Configurable option bits for this package """
    service = None
    service_name = None
    agent_host = None
    agent_port = None
    log_level = logging.WARN
    debug = None

    AGENT_DEFAULT_HOST = "localhost"
    AGENT_DEFAULT_PORT = 42699

    def __init__(self, **kwds):
        if "INSTANA_DEBUG" in os.environ:
            self.log_level = logging.DEBUG
            self.debug = True

        self.service_name = determine_service_name()
        self.agent_host = os.environ.get("INSTANA_AGENT_HOST", self.AGENT_DEFAULT_HOST)
        self.agent_port = os.environ.get("INSTANA_AGENT_PORT", self.AGENT_DEFAULT_PORT)

        if type(self.agent_port) is str:
            self.agent_port = int(self.agent_port)

        self.debug = os.environ.get("INSTANA_DEBUG", False)
        self.__dict__.update(kwds)


class AWSLambdaOptions:
    endpoint_url = None
    agent_key = None
    extra_http_headers = None
    timeout = None
    log_level = logging.WARN
    debug = None

    def __init__(self, **kwds):
        if "INSTANA_DEBUG" in os.environ:
            self.log_level = logging.DEBUG
            self.debug = True

        self.endpoint_url = os.environ.get("INSTANA_ENDPOINT_URL", None)

        # Remove any trailing slash (if any)
        if self.endpoint_url is not None and self.endpoint_url[-1] == "/":
            self.endpoint_url = self.endpoint_url[:-1]

        self.agent_key = os.environ.get("INSTANA_AGENT_KEY", None)

        self.extra_http_headers = os.environ.get("INSTANA_EXTRA_HTTP_HEADERS", None)
        self.timeout = os.environ.get("INSTANA_TIMEOUT", 0.5)
        self.log_level = os.environ.get("INSTANA_LOG_LEVEL", None)

        self.__dict__.update(kwds)



""" The in-process Instana agent that manages monitoring state and reporting that data. """
from __future__ import absolute_import

import json
import os
import time
from datetime import datetime
import threading
import requests

import instana.singletons
from instana.collector import Collector

from .fsm import TheMachine
from .log import logger
from .sensor import Sensor
from .util import to_json, get_py_source, package_version
from .options import StandardOptions, AWSLambdaOptions


class AnnounceData(object):
    """ The Announce Payload """
    pid = 0
    agentUuid = ""

    def __init__(self, **kwds):
        self.__dict__.update(kwds)


class AWSLambdaFrom(object):
    """ The source identifier for AWSLambdaAgent """
    hl = True
    cp = "aws"
    e = "qualifiedARN"

    def __init__(self, **kwds):
        self.__dict__.update(kwds)


class BaseAgent(object):
    """ Base class for all agent flavors """
    client = requests.Session()
    sensor = None

    def __init__(self):
        pass


class StandardAgent(BaseAgent):
    """
    The Agent class is the central controlling entity for the Instana Python language sensor.  The key
    parts it handles are the announce state and the collection and reporting of metrics and spans to the
    Instana Host agent.

    To do this, there are 3 major components to this class:
      1. TheMachine - finite state machine related to announce state
      2. Sensor -> Meter - metric collection and reporting
      3. Tracer -> Recorder - span queueing and reporting
    """
    AGENT_DISCOVERY_PATH = "com.instana.plugin.python.discovery"
    AGENT_DATA_PATH = "com.instana.plugin.python.%d"
    AGENT_HEADER = "Instana Agent"

    announce_data = None
    options = StandardOptions()

    machine = None
    last_seen = None
    last_fork_check = None
    _boot_pid = os.getpid()
    extra_headers = None
    secrets_matcher = 'contains-ignore-case'
    secrets_list = ['key', 'password', 'secret']
    should_threads_shutdown = threading.Event()

    def __init__(self):
        super(StandardAgent, self).__init__()
        logger.debug("initializing agent")
        self.sensor = Sensor(self)
        self.machine = TheMachine(self)

    def start(self, _):
        """
        Starts the agent and required threads

        This method is called after a successful announce.  See fsm.py
        """
        logger.debug("Spawning metric & span reporting threads")
        self.should_threads_shutdown.clear()
        self.sensor.start()
        instana.singletons.tracer.recorder.start()

    def handle_fork(self):
        """
        Forks happen.  Here we handle them.  Affected components are the singletons: Agent, Sensor & Tracers
        """
        # Reset the Agent
        self.reset()

    def reset(self):
        """
        This will reset the agent to a fresh unannounced state.
        :return: None
        """
        # Will signal to any running background threads to shutdown.
        self.should_threads_shutdown.set()

        self.last_seen = None
        self.announce_data = None

        # Will schedule a restart of the announce cycle in the future
        self.machine.reset()

    def is_timed_out(self):
        """
        If we haven't heard from the Instana host agent in 60 seconds, this
        method will return True.
        @return: Boolean
        """
        if self.last_seen and self.can_send:
            diff = datetime.now() - self.last_seen
            if diff.seconds > 60:
                return True
        return False

    def can_send(self):
        """
        Are we in a state where we can send data?
        @return: Boolean
        """
        # Watch for pid change (fork)
        current_pid = os.getpid()
        if self._boot_pid != current_pid:
            self._boot_pid = current_pid
            logger.debug("Fork detected; Handling like a pro...")
            self.handle_fork()
            return False

        if self.machine.fsm.current == "good2go":
            return True

        return False

    def set_from(self, json_string):
        """
        Sets the source identifiers given to use by the Instana Host agent.
        @param json_string: source identifiers
        @return: None
        """
        if type(json_string) is bytes:
            raw_json = json_string.decode("UTF-8")
        else:
            raw_json = json_string

        res_data = json.loads(raw_json)

        if "secrets" in res_data:
            self.secrets_matcher = res_data['secrets']['matcher']
            self.secrets_list = res_data['secrets']['list']

        if "extraHeaders" in res_data:
            self.extra_headers = res_data['extraHeaders']
            logger.info("Will also capture these custom headers: %s", self.extra_headers)

        self.announce_data = AnnounceData(pid=res_data['pid'], agentUuid=res_data['agentUuid'])

    def get_from_structure(self):
        """
        Retrieves the From data that is reported alongside monitoring data.
        @return: dict()
        """
        if os.environ.get("INSTANA_TEST", False):
            from_data = {'e': os.getpid(), 'h': 'fake'}
        else:
            from_data = {'e': self.announce_data.pid, 'h': self.announce_data.agentUuid}
        return from_data

    def is_agent_listening(self, host, port):
        """
        Check if the Instana Agent is listening on <host> and <port>.
        @return: Boolean
        """
        result = False
        try:
            url = "http://%s:%s/" % (host, port)
            response = self.client.get(url, timeout=0.8)

            server_header = response.headers["Server"]
            if server_header == self.AGENT_HEADER:
                logger.debug("Instana host agent found on %s:%d", host, port)
                result = True
            else:
                logger.debug("...something is listening on %s:%d but it's not the Instana Host Agent: %s",
                             host, port, server_header)
        except:
            logger.debug("Instana Host Agent not found on %s:%d", host, port)
        finally:
            return result

    def announce(self, discovery):
        """
        With the passed in Discovery class, attempt to announce to the host agent.
        """
        response = None
        try:
            url = self.__discovery_url()
            # logger.debug("making announce request to %s", url)
            response = self.client.put(url,
                                       data=to_json(discovery),
                                       headers={"Content-Type": "application/json"},
                                       timeout=0.8)

            if response.status_code == 200:
                self.last_seen = datetime.now()
        except Exception as e:
            logger.debug("announce: connection error (%s)", type(e))
        finally:
            return response

    def is_agent_ready(self):
        """
        Used after making a successful announce to test when the agent is ready to accept data.
        """
        ready = False
        try:
            response = self.client.head(self.__data_url(), timeout=0.8)

            if response.status_code == 200:
                ready = True
        except Exception as e:
            logger.debug("is_agent_ready: connection error (%s)", type(e))
        finally:
            return ready

    def report_data_payload(self, entity_data):
        """
        Used to report entity data (metrics & snapshot) to the host agent.
        """
        response = None
        try:
            response = self.client.post(self.__data_url(),
                                        data=to_json(entity_data),
                                        headers={"Content-Type": "application/json"},
                                        timeout=0.8)

            # logger.warning("report_data: response.status_code is %s" % response.status_code)

            if response.status_code == 200:
                self.last_seen = datetime.now()
        except Exception as e:
            logger.debug("report_data_payload: Instana host agent connection error (%s)", type(e))
        finally:
            return response

    def report_traces(self, spans):
        """
        Used to report entity data (metrics & snapshot) to the host agent.
        """
        response = None
        try:
            # Concurrency double check:  Don't report if we don't have
            # any spans
            if len(spans) == 0:
                return 0

            response = self.client.post(self.__traces_url(),
                                        data=to_json(spans),
                                        headers={"Content-Type": "application/json"},
                                        timeout=0.8)

            # logger.debug("report_traces: response.status_code is %s" % response.status_code)

            if response.status_code == 200:
                self.last_seen = datetime.now()
        except Exception as e:
            logger.debug("report_traces: Instana host agent connection error (%s)", type(e))
        finally:
            return response

    def handle_agent_tasks(self, task):
        """
        When request(s) are received by the host agent, it is sent here
        for handling & processing.
        """
        logger.debug("Received agent request with messageId: %s", task["messageId"])
        if "action" in task:
            if task["action"] == "python.source":
                payload = get_py_source(task["args"]["file"])
            else:
                message = "Unrecognized action: %s. An newer Instana package may be required " \
                          "for this. Current version: %s" % (task["action"], package_version())
                payload = {"error": message}
        else:
            payload = {"error": "Instana Python: No action specified in request."}

        self.__task_response(task["messageId"], payload)

    def __task_response(self, message_id, data):
        """
        When the host agent passes us a task and we do it, this function is used to
        respond with the results of the task.
        """
        response = None
        try:
            payload = json.dumps(data)

            logger.debug("Task response is %s: %s", self.__response_url(message_id), payload)

            response = self.client.post(self.__response_url(message_id),
                                        data=payload,
                                        headers={"Content-Type": "application/json"},
                                        timeout=0.8)
        except Exception as e:
            logger.debug("__task_response: Instana host agent connection error (%s)", type(e))
        finally:
            return response

    def __discovery_url(self):
        """
        URL for announcing to the host agent
        """
        return "http://%s:%s/%s" % (self.options.agent_host, self.options.agent_port, self.AGENT_DISCOVERY_PATH)

    def __data_url(self):
        """
        URL for posting metrics to the host agent.  Only valid when announced.
        """
        path = self.AGENT_DATA_PATH % self.announce_data.pid
        return "http://%s:%s/%s" % (self.options.agent_host, self.options.agent_port, path)

    def __traces_url(self):
        """
        URL for posting traces to the host agent.  Only valid when announced.
        """
        path = "com.instana.plugin.python/traces.%d" % self.announce_data.pid
        return "http://%s:%s/%s" % (self.options.agent_host, self.options.agent_port, path)

    def __response_url(self, message_id):
        """
        URL for responding to agent requests.
        """
        path = "com.instana.plugin.python/response.%d?messageId=%s" % (int(self.announce_data.pid), message_id)
        return "http://%s:%s/%s" % (self.options.agent_host, self.options.agent_port, path)


class AWSLambdaAgent(BaseAgent):
    """ In-process agent for AWS Lambda """
    def __init__(self):
        super(AWSLambdaAgent, self).__init__()

        self.from_ = AWSLambdaFrom()
        self.collector = None
        self.options = AWSLambdaOptions()
        self.report_headers = None
        self._can_send = False
        self.extra_headers = None

        if "INSTANA_EXTRA_HTTP_HEADERS" in os.environ:
            self.extra_headers = str(os.environ["INSTANA_EXTRA_HTTP_HEADERS"]).lower().split(';')

        if self._validate_options():
            self._can_send = True
            self.collector = Collector(self)
            self.collector.start()
        else:
            logger.warning("Required INSTANA_AGENT_KEY and/or INSTANA_ENDPOINT_URL environment variables not set.  "
                           "We will not be able monitor this function.")

    def can_send(self):
        """
        Are we in a state where we can send data?
        @return: Boolean
        """
        return self._can_send

    def get_from_structure(self):
        """
        Retrieves the From data that is reported alongside monitoring data.
        @return: dict()
        """
        return {'hl': True, 'cp': 'aws', 'e': self.collector.context.invoked_function_arn}

    def report_data_payload(self, payload):
        """
        Used to report metrics and span data to the endpoint URL in self.options.endpoint_url
        """
        response = None
        try:
            if self.report_headers is None:
                # Prepare request headers
                self.report_headers = dict()
                self.report_headers["Content-Type"] = "application/json"
                self.report_headers["X-Instana-Host"] = self.collector.context.invoked_function_arn
                self.report_headers["X-Instana-Key"] = self.options.agent_key
                self.report_headers["X-Instana-Time"] = str(round(time.time() * 1000))

            logger.debug("using these headers: %s", self.report_headers)

            if 'INSTANA_DISABLE_CA_CHECK' in os.environ:
                ssl_verify = False
            else:
                ssl_verify = True

            response = self.client.post(self.__data_bundle_url(),
                                        data=to_json(payload),
                                        headers=self.report_headers,
                                        timeout=self.options.timeout,
                                        verify=ssl_verify)

            logger.debug("report_data_payload: response.status_code is %s", response.status_code)
        except Exception as e:
            logger.debug("report_data_payload: connection error (%s)", type(e))
        finally:
            return response

    def _validate_options(self):
        """
        Validate that the options used by this Agent are valid.  e.g. can we report data?
        """
        return self.options.endpoint_url is not None and self.options.agent_key is not None

    def __data_bundle_url(self):
        """
        URL for posting metrics to the host agent.  Only valid when announced.
        """
        return "%s/bundle" % self.options.endpoint_url



import os
import sys
import opentracing

from .agent import StandardAgent, AWSLambdaAgent
from .log import logger
from .tracer import InstanaTracer
from .recorder import StandardRecorder, AWSLambdaRecorder

agent = None
tracer = None
span_recorder = None

if os.environ.get("INSTANA_ENDPOINT_URL", False):
    agent = AWSLambdaAgent()
    span_recorder = AWSLambdaRecorder(agent)
else:
    agent = StandardAgent()
    span_recorder = StandardRecorder()


def get_agent():
    """
    Retrieve the globally configured agent
    @return: The Instana Agent singleton
    """
    global agent
    return agent


def set_agent(new_agent):
    """
    Set the global agent for the Instana package.  This is used for the
    test suite only currently.

    @param new_agent: agent to replace current singleton
    @return: None
    """
    global agent
    agent = new_agent


# The global OpenTracing compatible tracer used internally by
# this package.
tracer = InstanaTracer(recorder=span_recorder)

if sys.version_info >= (3, 4):
    try:
        from opentracing.scope_managers.asyncio import AsyncioScopeManager
        async_tracer = InstanaTracer(scope_manager=AsyncioScopeManager(), recorder=span_recorder)
    except Exception:
        logger.debug("Error setting up async_tracer:", exc_info=True)


# Mock the tornado tracer until tornado is detected and instrumented first
tornado_tracer = tracer


def setup_tornado_tracer():
    global tornado_tracer
    from opentracing.scope_managers.tornado import TornadoScopeManager
    tornado_tracer = InstanaTracer(scope_manager=TornadoScopeManager(), recorder=span_recorder)


# Set ourselves as the tracer.
opentracing.tracer = tracer


def get_tracer():
    """
    Retrieve the globally configured tracer
    @return: Tracer
    """
    global tracer
    return tracer


def set_tracer(new_tracer):
    """
    Set the global tracer for the Instana package.  This is used for the
    test suite only currently.
    @param new_tracer: The new tracer to replace the singleton
    @return: None
    """
    global tracer
    tracer = new_tracer



from .log import logger
from .util import DictionaryOfStan
from basictracer.span import BasicSpan
import opentracing.ext.tags as ot_tags


class SpanContext():
    def __init__(
            self,
            trace_id=None,
            span_id=None,
            baggage=None,
            sampled=True,
            level=1):

        self.level = level
        self.trace_id = trace_id
        self.span_id = span_id
        self.sampled = sampled
        self._baggage = baggage or {}

    @property
    def baggage(self):
        return self._baggage

    def with_baggage_item(self, key, value):
        new_baggage = self._baggage.copy()
        new_baggage[key] = value
        return SpanContext(
            trace_id=self.trace_id,
            span_id=self.span_id,
            sampled=self.sampled,
            baggage=new_baggage)


class InstanaSpan(BasicSpan):
    stack = None

    def finish(self, finish_time=None):
        super(InstanaSpan, self).finish(finish_time)

    def mark_as_errored(self, tags = None):
        """
        Mark this span as errored.

        @param tags: optional tags to add to the span
        """
        try:
            ec = self.tags.get('ec', 0)
            self.set_tag('ec', ec + 1)

            if tags is not None and type(tags) is dict:
                for key in tags:
                    self.set_tag(key, tags[key])
        except Exception:
            logger.debug('span.mark_as_errored', exc_info=True)

    def assure_errored(self):
        """
        Make sure that this span is marked as errored.
        @return: None
        """
        try:
            ec = self.tags.get('ec', None)
            if ec is None or ec == 0:
                self.set_tag('ec', 1)
        except Exception:
            logger.debug('span.assure_errored', exc_info=True)

    def log_exception(self, e):
        """
        Log an exception onto this span.  This will log pertinent info from the exception and
        assure that this span is marked as errored.

        @param e: the exception to log
        """
        try:
            message = ""
            self.mark_as_errored()

            if hasattr(e, '__str__') and len(str(e)) > 0:
                message = str(e)
            elif hasattr(e, 'message') and e.message is not None:
                message = e.message
            else:
                message = repr(e)

            if self.operation_name in ['rpc-server', 'rpc-client']:
                self.set_tag('rpc.error', message)
            elif self.operation_name == "mysql":
                self.set_tag('mysql.error', message)
            elif self.operation_name == "postgres":
                self.set_tag('pg.error', message)
            elif self.operation_name in RegisteredSpan.HTTP_SPANS:
                self.set_tag('http.error', message)
            else:
                self.log_kv({'message': message})
        except Exception:
            logger.debug("span.log_exception", exc_info=True)
            raise

    def collect_logs(self):
        """
        Collect up log data and feed it to the Instana brain.

        :param span: The span to search for logs in
        :return: Logs ready for consumption by the Instana brain.
        """
        logs = {}
        for log in self.logs:
            ts = int(round(log.timestamp * 1000))
            if ts not in logs:
                logs[ts] = {}

            if 'message' in log.key_values:
                logs[ts]['message'] = log.key_values['message']
            if 'event' in log.key_values:
                logs[ts]['event'] = log.key_values['event']
            if 'parameters' in log.key_values:
                logs[ts]['parameters'] = log.key_values['parameters']

        return logs


class BaseSpan(object):
    def __str__(self):
        return "BaseSpan(%s)" % self.__dict__.__str__()

    def __repr__(self):
        return self.__dict__.__str__()

    def __init__(self, span, source, **kwargs):
        self.t = span.context.trace_id
        self.p = span.parent_id
        self.s = span.context.span_id
        self.ts = int(round(span.start_time * 1000))
        self.d = int(round(span.duration * 1000))
        self.f = source
        self.ec = span.tags.pop('ec', None)

        if span.stack:
            self.stack = span.stack

        self.__dict__.update(kwargs)


class SDKSpan(BaseSpan):
    ENTRY_KIND = ["entry", "server", "consumer"]
    EXIT_KIND = ["exit", "client", "producer"]

    def __init__(self, span, source, service_name, **kwargs):
        super(SDKSpan, self).__init__(span, source, **kwargs)
        self.n = "sdk"
        self.k = self.get_span_kind_as_int(span)

        self.data = DictionaryOfStan()
        self.data["sdk"]["name"] = span.operation_name
        self.data["sdk"]["type"] = self.get_span_kind_as_string(span)
        self.data["sdk"]["custom"]["tags"] = span.tags
        self.data["sdk"]["custom"]["logs"] = span.logs
        self.data["service"] = service_name

        # self.data = Data()
        # self.data.sdk = SDKData(name=span.operation_name, Type=self.get_span_kind_as_string(span))
        # self.data.sdk.custom = CustomData(tags=span.tags, logs=span.collect_logs())
        # self.data.service = service_name

        if "arguments" in span.tags:
            self.data.sdk.arguments = span.tags["arguments"]

        if "return" in span.tags:
            self.data.sdk.Return = span.tags["return"]

        if len(span.context.baggage) > 0:
            self.data["baggage"] = span.context.baggage

    def get_span_kind_as_string(self, span):
        """
            Will retrieve the `span.kind` tag and return the appropriate string value for the Instana backend or
            None if the tag is set to something we don't recognize.

        :param span: The span to search for the `span.kind` tag
        :return: String
        """
        kind = None
        if "span.kind" in span.tags:
            if span.tags["span.kind"] in self.ENTRY_KIND:
                kind = "entry"
            elif span.tags["span.kind"] in self.EXIT_KIND:
                kind = "exit"
            else:
                kind = "intermediate"
        return kind

    def get_span_kind_as_int(self, span):
        """
            Will retrieve the `span.kind` tag and return the appropriate integer value for the Instana backend or
            None if the tag is set to something we don't recognize.

        :param span: The span to search for the `span.kind` tag
        :return: Integer
        """
        kind = None
        if "span.kind" in span.tags:
            if span.tags["span.kind"] in self.ENTRY_KIND:
                kind = 1
            elif span.tags["span.kind"] in self.EXIT_KIND:
                kind = 2
            else:
                kind = 3
        return kind


class RegisteredSpan(BaseSpan):
    HTTP_SPANS = ("aiohttp-client", "aiohttp-server", "django", "http", "soap", "tornado-client",
                  "tornado-server", "urllib3", "wsgi")

    EXIT_SPANS = ("aiohttp-client", "cassandra", "couchbase", "log", "memcache", "mongo", "mysql", "postgres",
                  "rabbitmq", "redis", "rpc-client", "sqlalchemy", "soap", "tornado-client", "urllib3",
                  "pymongo")

    ENTRY_SPANS = ("aiohttp-server", "aws.lambda.entry", "django", "wsgi", "rabbitmq", "rpc-server", "tornado-server")

    LOCAL_SPANS = ("render")

    def __init__(self, span, source, **kwargs):
        super(RegisteredSpan, self).__init__(span, source, **kwargs)
        self.n = span.operation_name
        self.data = DictionaryOfStan()

        self.k = 1
        if span.operation_name in self.ENTRY_SPANS:
            # entry
            self._populate_entry_span_data(span)
        elif span.operation_name in self.EXIT_SPANS:
            self.k = 2 # exit
            self._populate_exit_span_data(span)
        elif span.operation_name in self.LOCAL_SPANS:
            self.k = 3 # intermediate span
            self._populate_local_span_data(span)

        if "rabbitmq" in self.data and self.data["rabbitmq"]["sort"] == "consume":
            self.k = 1  # entry

        # Store any leftover tags in the custom section
        if len(span.tags):
            self.data["custom"]["tags"] = span.tags

    def _populate_entry_span_data(self, span):
        if span.operation_name in self.HTTP_SPANS:
            self._collect_http_tags(span)

        elif span.operation_name == "aws.lambda.entry":
            self.data["lambda"]["arn"] = span.tags.pop('lambda.arn', "Unknown")
            self.data["lambda"]["alias"] = None
            self.data["lambda"]["runtime"] = "python"
            self.data["lambda"]["functionName"] = span.tags.pop('lambda.name', "Unknown")
            self.data["lambda"]["functionVersion"] = span.tags.pop('lambda.version', "Unknown")
            self.data["lambda"]["trigger"] = span.tags.pop('lambda.trigger', None)
            self.data["lambda"]["error"] = None

            trigger_type = self.data["lambda"]["trigger"]

            if trigger_type in ["aws:api.gateway", "aws:application.load.balancer"]:
                self._collect_http_tags(span)
            elif trigger_type == 'aws:cloudwatch.events':
                self.data["lambda"]["cw"]["events"]["id"] = span.tags.pop('data.lambda.cw.events.id', None)
                self.data["lambda"]["cw"]["events"]["more"] = span.tags.pop('lambda.cw.events.more', False)
                self.data["lambda"]["cw"]["events"]["resources"] = span.tags.pop('lambda.cw.events.resources', None)

            elif trigger_type == 'aws:cloudwatch.logs':
                self.data["lambda"]["cw"]["logs"]["group"] = span.tags.pop('lambda.cw.logs.group', None)
                self.data["lambda"]["cw"]["logs"]["stream"] = span.tags.pop('lambda.cw.logs.stream', None)
                self.data["lambda"]["cw"]["logs"]["more"] = span.tags.pop('lambda.cw.logs.more', None)
                self.data["lambda"]["cw"]["logs"]["events"] = span.tags.pop('lambda.cw.logs.events', None)

            elif trigger_type == 'aws:s3':
                self.data["lambda"]["s3"]["events"] = span.tags.pop('lambda.s3.events', None)
            elif trigger_type == 'aws:sqs':
                self.data["lambda"]["sqs"]["messages"] = span.tags.pop('lambda.sqs.messages', None)

        elif span.operation_name == "rabbitmq":
            self.data["rabbitmq"]["exchange"] = span.tags.pop('exchange', None)
            self.data["rabbitmq"]["queue"] = span.tags.pop('queue', None)
            self.data["rabbitmq"]["sort"] = span.tags.pop('sort', None)
            self.data["rabbitmq"]["address"] = span.tags.pop('address', None)
            self.data["rabbitmq"]["key"] = span.tags.pop('key', None)

        elif span.operation_name == "rpc-server":
            self.data["rpc"]["flavor"] = span.tags.pop('rpc.flavor', None)
            self.data["rpc"]["host"] = span.tags.pop('rpc.host', None)
            self.data["rpc"]["port"] = span.tags.pop('rpc.port', None)
            self.data["rpc"]["call"] = span.tags.pop('rpc.call', None)
            self.data["rpc"]["call_type"] = span.tags.pop('rpc.call_type', None)
            self.data["rpc"]["params"] = span.tags.pop('rpc.params', None)
            self.data["rpc"]["baggage"] = span.tags.pop('rpc.baggage', None)
            self.data["rpc"]["error"] = span.tags.pop('rpc.error', None)
        else:
            logger.debug("SpanRecorder: Unknown entry span: %s" % span.operation_name)

    def _populate_local_span_data(self, span):
        if span.operation_name == "render":
            self.data["render"]["name"] = span.tags.pop('name', None)
            self.data["render"]["type"] = span.tags.pop('type', None)
            self.data["log"]["message"] = span.tags.pop('message', None)
            self.data["log"]["parameters"] = span.tags.pop('parameters', None)
        else:
            logger.debug("SpanRecorder: Unknown local span: %s" % span.operation_name)

    def _populate_exit_span_data(self, span):
        if span.operation_name in self.HTTP_SPANS:
            self._collect_http_tags(span)
        elif span.operation_name == "rabbitmq":
            self.data["rabbitmq"]["exchange"] = span.tags.pop('exchange', None)
            self.data["rabbitmq"]["queue"] = span.tags.pop('queue', None)
            self.data["rabbitmq"]["sort"] = span.tags.pop('sort', None)
            self.data["rabbitmq"]["address"] = span.tags.pop('address', None)
            self.data["rabbitmq"]["key"] = span.tags.pop('key', None)

        elif span.operation_name == "cassandra":
            self.data["cassandra"]["cluster"] = span.tags.pop('cassandra.cluster', None)
            self.data["cassandra"]["query"] = span.tags.pop('cassandra.query', None)
            self.data["cassandra"]["keyspace"] = span.tags.pop('cassandra.keyspace', None)
            self.data["cassandra"]["fetchSize"] = span.tags.pop('cassandra.fetchSize', None)
            self.data["cassandra"]["achievedConsistency"] = span.tags.pop('cassandra.achievedConsistency', None)
            self.data["cassandra"]["triedHosts"] = span.tags.pop('cassandra.triedHosts', None)
            self.data["cassandra"]["fullyFetched"] = span.tags.pop('cassandra.fullyFetched', None)
            self.data["cassandra"]["error"] = span.tags.pop('cassandra.error', None)

        elif span.operation_name == "couchbase":
            self.data["couchbase"]["hostname"] = span.tags.pop('couchbase.hostname', None)
            self.data["couchbase"]["bucket"] = span.tags.pop('couchbase.bucket', None)
            self.data["couchbase"]["type"] = span.tags.pop('couchbase.type', None)
            self.data["couchbase"]["error"] = span.tags.pop('couchbase.error', None)
            self.data["couchbase"]["error_type"] = span.tags.pop('couchbase.error_type', None)
            self.data["couchbase"]["sql"] = span.tags.pop('couchbase.sql', None)

        elif span.operation_name == "redis":
            self.data["redis"]["connection"] = span.tags.pop('connection', None)
            self.data["redis"]["driver"] = span.tags.pop('driver', None)
            self.data["redis"]["command"] = span.tags.pop('command', None)
            self.data["redis"]["error"] = span.tags.pop('redis.error', None)
            self.data["redis"]["subCommands"] = span.tags.pop('subCommands', None)

        elif span.operation_name == "rpc-client":
            self.data["rpc"]["flavor"] = span.tags.pop('rpc.flavor', None)
            self.data["rpc"]["host"] = span.tags.pop('rpc.host', None)
            self.data["rpc"]["port"] = span.tags.pop('rpc.port', None)
            self.data["rpc"]["call"] = span.tags.pop('rpc.call', None)
            self.data["rpc"]["call_type"] = span.tags.pop('rpc.call_type', None)
            self.data["rpc"]["params"] = span.tags.pop('rpc.params', None)
            self.data["rpc"]["baggage"] = span.tags.pop('rpc.baggage', None)
            self.data["rpc"]["error"] = span.tags.pop('rpc.error', None)

        elif span.operation_name == "sqlalchemy":
            self.data["sqlalchemy"]["sql"] = span.tags.pop('sqlalchemy.sql', None)
            self.data["sqlalchemy"]["eng"] = span.tags.pop('sqlalchemy.eng', None)
            self.data["sqlalchemy"]["url"] = span.tags.pop('sqlalchemy.url', None)
            self.data["sqlalchemy"]["err"] = span.tags.pop('sqlalchemy.err', None)

        elif span.operation_name == "mysql":
            self.data["mysql"]["host"] = span.tags.pop('host', None)
            self.data["mysql"]["port"] = span.tags.pop('port', None)
            self.data["mysql"]["db"] = span.tags.pop(ot_tags.DATABASE_INSTANCE, None)
            self.data["mysql"]["user"] = span.tags.pop(ot_tags.DATABASE_USER, None)
            self.data["mysql"]["stmt"] = span.tags.pop(ot_tags.DATABASE_STATEMENT, None)
            self.data["mysql"]["error"] = span.tags.pop('mysql.error', None)

        elif span.operation_name == "postgres":
            self.data["pg"]["host"] = span.tags.pop('host', None)
            self.data["pg"]["port"] = span.tags.pop('port', None)
            self.data["pg"]["db"] = span.tags.pop(ot_tags.DATABASE_INSTANCE, None)
            self.data["pg"]["user"] = span.tags.pop(ot_tags.DATABASE_USER, None)
            self.data["pg"]["stmt"] = span.tags.pop(ot_tags.DATABASE_STATEMENT, None)
            self.data["pg"]["error"] = span.tags.pop('pg.error', None)

        elif span.operation_name == "mongo":
            service = "%s:%s" % (span.tags.pop('host', None), span.tags.pop('port', None))
            namespace = "%s.%s" % (span.tags.pop('db', "?"), span.tags.pop('collection', "?"))

            self.data["mongo"]["service"] = service
            self.data["mongo"]["namespace"] = namespace
            self.data["mongo"]["command"] = span.tags.pop('command', None)
            self.data["mongo"]["filter"] = span.tags.pop('filter', None)
            self.data["mongo"]["json"] = span.tags.pop('json', None)
            self.data["mongo"]["error"] = span.tags.pop('error', None)

        elif span.operation_name == "log":
            # use last special key values
            for l in span.logs:
                if "message" in l.key_values:
                    self.data["log"]["message"] = l.key_values.pop("message", None)
                if "parameters" in l.key_values:
                    self.data["log"]["parameters"] = l.key_values.pop("parameters", None)
        else:
            logger.debug("SpanRecorder: Unknown exit span: %s" % span.operation_name)

    def _collect_http_tags(self, span):
        self.data["http"]["host"] = span.tags.pop("http.host", None)
        self.data["http"]["url"] = span.tags.pop(ot_tags.HTTP_URL, None)
        self.data["http"]["path"] = span.tags.pop("http.path", None)
        self.data["http"]["params"] = span.tags.pop('http.params', None)
        self.data["http"]["method"] = span.tags.pop(ot_tags.HTTP_METHOD, None)
        self.data["http"]["status"] = span.tags.pop(ot_tags.HTTP_STATUS_CODE, None)
        self.data["http"]["path_tpl"] = span.tags.pop("http.path_tpl", None)
        self.data["http"]["error"] = span.tags.pop('http.error', None)

        if span.operation_name == "soap":
            self.data["soap"]["action"] = span.tags.pop('soap.action', None)



from __future__ import absolute_import

import re

from ..log import logger
from ..singletons import tracer

try:
    import sqlalchemy
    from sqlalchemy import event
    from sqlalchemy.engine import Engine

    url_regexp = re.compile(r"\/\/(\S+@)")

    @event.listens_for(Engine, 'before_cursor_execute', named=True)
    def receive_before_cursor_execute(**kw):
        try:
            parent_span = tracer.active_span

            # If we're not tracing, just return
            if parent_span is None:
                return

            scope = tracer.start_active_span("sqlalchemy", child_of=parent_span)
            context = kw['context']
            context._stan_scope = scope

            conn = kw['conn']
            url = str(conn.engine.url)
            scope.span.set_tag('sqlalchemy.sql', kw['statement'])
            scope.span.set_tag('sqlalchemy.eng', conn.engine.name)
            scope.span.set_tag('sqlalchemy.url', url_regexp.sub('//', url))
        except Exception as e:
            logger.debug(e)
        finally:
            return

    @event.listens_for(Engine, 'after_cursor_execute', named=True)
    def receive_after_cursor_execute(**kw):
        context = kw['context']

        if context is not None and hasattr(context, '_stan_scope'):
            scope = context._stan_scope
            if scope is not None:
                scope.close()

    @event.listens_for(Engine, 'dbapi_error', named=True)
    def receive_dbapi_error(**kw):
        context = kw['context']

        if context is not None and hasattr(context, '_stan_scope'):
            scope = context._stan_scope
            if scope is not None:
                scope.span.mark_as_errored()

                if 'exception' in kw:
                    e = kw['exception']
                    scope.span.set_tag('sqlalchemy.err', str(e))
                else:
                    scope.span.set_tag('sqlalchemy.err', "No dbapi error specified.")
                scope.close()


    logger.debug("Instrumenting sqlalchemy")
except ImportError:
    pass



from __future__ import absolute_import

import wrapt

from ...singletons import tracer


@wrapt.patch_function_wrapper('flask', 'templating._render')
def render_with_instana(wrapped, instance, argv, kwargs):
    ctx = argv[1]

    # If we're not tracing, just return
    if not hasattr(ctx['g'], 'scope'):
        return wrapped(*argv, **kwargs)

    with tracer.start_active_span("render", child_of=ctx['g'].scope.span) as rscope:
        try:
            template = argv[0]

            rscope.span.set_tag("type", "template")
            if template.name is None:
                rscope.span.set_tag("name", '(from string)')
            else:
                rscope.span.set_tag("name", template.name)

            return wrapped(*argv, **kwargs)
        except Exception as e:
            rscope.span.log_exception(e)
            raise



from __future__ import absolute_import

import opentracing
import wrapt

from ...log import logger
from ...singletons import agent, setup_tornado_tracer, tornado_tracer
from ...util import strip_secrets

from distutils.version import LooseVersion

try:
    import tornado
    from opentracing.scope_managers.tornado import tracer_stack_context

    setup_tornado_tracer()

    # Tornado >=6.0 switched to contextvars for context management.  This requires changes to the opentracing
    # scope managers which we will tackle soon.
    # Limit Tornado version for the time being.
    if hasattr(tornado, 'version') and (LooseVersion(tornado.version) < LooseVersion('6.0.0')):

        @wrapt.patch_function_wrapper('tornado.web', 'RequestHandler._execute')
        def execute_with_instana(wrapped, instance, argv, kwargs):
            try:
                with tracer_stack_context():
                    ctx = tornado_tracer.extract(opentracing.Format.HTTP_HEADERS, instance.request.headers)
                    scope = tornado_tracer.start_active_span('tornado-server', child_of=ctx)

                    # Query param scrubbing
                    if instance.request.query is not None and len(instance.request.query) > 0:
                        cleaned_qp = strip_secrets(instance.request.query, agent.secrets_matcher, agent.secrets_list)
                        scope.span.set_tag("http.params", cleaned_qp)

                    url = "%s://%s%s" % (instance.request.protocol, instance.request.host, instance.request.path)
                    scope.span.set_tag("http.url", url)
                    scope.span.set_tag("http.method", instance.request.method)

                    scope.span.set_tag("handler", instance.__class__.__name__)

                    # Custom header tracking support
                    if hasattr(agent, 'extra_headers') and agent.extra_headers is not None:
                        for custom_header in agent.extra_headers:
                            if custom_header in instance.request.headers:
                                scope.span.set_tag("http.%s" % custom_header, instance.request.headers[custom_header])

                    setattr(instance.request, "_instana", scope)

                    # Set the context response headers now because tornado doesn't give us a better option to do so
                    # later for this request.
                    tornado_tracer.inject(scope.span.context, opentracing.Format.HTTP_HEADERS, instance._headers)
                    instance.set_header(name='Server-Timing', value="intid;desc=%s" % scope.span.context.trace_id)

                    return wrapped(*argv, **kwargs)
            except Exception:
                logger.debug("tornado execute", exc_info=True)


        @wrapt.patch_function_wrapper('tornado.web', 'RequestHandler.set_default_headers')
        def set_default_headers_with_instana(wrapped, instance, argv, kwargs):
            if not hasattr(instance.request, '_instana'):
                return wrapped(*argv, **kwargs)

            scope = instance.request._instana
            tornado_tracer.inject(scope.span.context, opentracing.Format.HTTP_HEADERS, instance._headers)
            instance.set_header(name='Server-Timing', value="intid;desc=%s" % scope.span.context.trace_id)


        @wrapt.patch_function_wrapper('tornado.web', 'RequestHandler.on_finish')
        def on_finish_with_instana(wrapped, instance, argv, kwargs):
            try:
                if not hasattr(instance.request, '_instana'):
                    return wrapped(*argv, **kwargs)

                scope = instance.request._instana
                status_code = instance.get_status()

                # Mark 500 responses as errored
                if 500 <= status_code <= 511:
                    scope.span.mark_as_errored()

                scope.span.set_tag("http.status_code", status_code)
                scope.close()

                return wrapped(*argv, **kwargs)
            except Exception:
                logger.debug("tornado on_finish", exc_info=True)

        @wrapt.patch_function_wrapper('tornado.web', 'RequestHandler.log_exception')
        def log_exception_with_instana(wrapped, instance, argv, kwargs):
            try:
                if not hasattr(instance.request, '_instana'):
                    return wrapped(*argv, **kwargs)

                if not isinstance(argv[1], tornado.web.HTTPError):
                    scope = instance.request._instana
                    scope.span.log_exception(argv[0])

                return wrapped(*argv, **kwargs)
            except Exception:
                logger.debug("tornado log_exception", exc_info=True)

        logger.debug("Instrumenting tornado server")
except ImportError:
    pass




from __future__ import absolute_import

from .meter import Meter


class Sensor(object):
    agent = None
    meter = None

    def __init__(self, agent):
        self.agent = agent
        self.meter = Meter(agent)

    def start(self):
        # Nothing to do for the Sensor;  Pass onto Meter
        self.meter.start()

    def handle_fork(self):
        # Nothing to do for the Sensor;  Pass onto Meter
        self.meter.handle_fork()



from __future__ import absolute_import

import opentracing
import wrapt

from ...log import logger
from ...singletons import agent, async_tracer
from ...util import strip_secrets


try:
    import aiohttp
    import asyncio

    from aiohttp.web import middleware

    @middleware
    async def stan_middleware(request, handler):
        try:
            ctx = async_tracer.extract(opentracing.Format.HTTP_HEADERS, request.headers)
            request['scope'] = async_tracer.start_active_span('aiohttp-server', child_of=ctx)
            scope = request['scope']

            # Query param scrubbing
            url = str(request.url)
            parts = url.split('?')
            if len(parts) > 1:
                cleaned_qp = strip_secrets(parts[1], agent.secrets_matcher, agent.secrets_list)
                scope.span.set_tag("http.params", cleaned_qp)

            scope.span.set_tag("http.url", parts[0])
            scope.span.set_tag("http.method", request.method)

            # Custom header tracking support
            if hasattr(agent, 'extra_headers') and agent.extra_headers is not None:
                for custom_header in agent.extra_headers:
                    if custom_header in request.headers:
                        scope.span.set_tag("http.%s" % custom_header, request.headers[custom_header])

            response = await handler(request)

            if response is not None:
                # Mark 500 responses as errored
                if 500 <= response.status <= 511:
                    scope.span.mark_as_errored()

                scope.span.set_tag("http.status_code", response.status)
                async_tracer.inject(scope.span.context, opentracing.Format.HTTP_HEADERS, response.headers)
                response.headers['Server-Timing'] = "intid;desc=%s" % scope.span.context.trace_id

            return response
        except Exception as e:
            logger.debug("aiohttp stan_middleware", exc_info=True)
            if scope is not None:
                scope.span.set_tag("http.status_code", 500)
                scope.span.log_exception(e)
            raise
        finally:
            if scope is not None:
                scope.close()


    @wrapt.patch_function_wrapper('aiohttp.web','Application.__init__')
    def init_with_instana(wrapped, instance, argv, kwargs):
        if "middlewares" in kwargs:
            kwargs["middlewares"].insert(0, stan_middleware)
        else:
            kwargs["middlewares"] = [stan_middleware]

        return wrapped(*argv, **kwargs)

    logger.debug("Instrumenting aiohttp server")
except ImportError:
    pass




import os
import sys
import threading

from .log import logger
from .util import every, DictionaryOfStan


if sys.version_info.major == 2:
    import Queue as queue
else:
    import queue


class Collector(object):
    def __init__(self, agent):
        logger.debug("Loading collector")
        self.agent = agent
        self.span_queue = queue.Queue()
        self.thread_shutdown = threading.Event()
        self.thread_shutdown.clear()
        self.context = None
        self.event = None
        self.snapshot_data = None
        self.snapshot_data_sent = False
        self.lock = threading.Lock()

    def start(self):
        if self.agent.can_send():
            t = threading.Thread(target=self.thread_loop, args=())
            t.setDaemon(True)
            t.start()
        else:
            logger.warn("Collector started but the agent tells us we can't send anything out.")

    def shutdown(self):
        logger.debug("Collector.shutdown: Reporting final data.")
        self.thread_shutdown.set()
        self.prepare_and_report_data()

    def thread_loop(self):
        every(5, self.background_report, "Instana Collector: prepare_and_report_data")

    def background_report(self):
        if self.thread_shutdown.is_set():
            logger.debug("Thread shutdown signal is active: Shutting down reporting thread")
            return False
        return self.prepare_and_report_data()

    def prepare_payload(self):
        payload = DictionaryOfStan()
        payload["spans"] = None
        payload["metrics"] = None

        if not self.span_queue.empty():
            payload["spans"] = self.__queued_spans()

        if self.snapshot_data and self.snapshot_data_sent is False:
            payload["metrics"] = self.snapshot_data
            self.snapshot_data_sent = True

        return payload

    def prepare_and_report_data(self):
        if "INSTANA_TEST" in os.environ:
            return True

        lock_acquired = self.lock.acquire(False)
        if lock_acquired:
            payload = self.prepare_payload()

            if len(payload) > 0:
                self.agent.report_data_payload(payload)
            else:
                logger.debug("prepare_and_report_data: No data to report")
            self.lock.release()
        else:
            logger.debug("prepare_and_report_data: Couldn't acquire lock")
        return True

    def collect_snapshot(self, event, context):
        self.snapshot_data = DictionaryOfStan()

        self.context = context
        self.event = event

        try:
            self.snapshot_data["plugins"]["name"] = "com.instana.plugin.aws.lambda"
            self.snapshot_data["plugins"]["entityId"] = self.context.invoked_function_arn
        except:
            logger.debug("collect_snapshot error", exc_info=True)
        finally:
            return self.snapshot_data

    def __queued_spans(self):
        """ Get all of the spans in the queue """
        span = None
        spans = []
        while True:
            try:
                span = self.span_queue.get(False)
            except queue.Empty:
                break
            else:
                spans.append(span)
        return spans



from __future__ import absolute_import

import os
import sys
import threading

from .log import logger
from .util import every
import instana.singletons
from basictracer import Sampler
from .span import (RegisteredSpan, SDKSpan)

if sys.version_info.major == 2:
    import Queue as queue
else:
    import queue


class StandardRecorder(object):
    THREAD_NAME = "Instana Span Reporting"

    REGISTERED_SPANS = ("aiohttp-client", "aiohttp-server", "aws.lambda.entry", "cassandra", "couchbase",
                        "django", "log", "memcache", "mongo", "mysql", "postgres", "pymongo", "rabbitmq", "redis",
                        "render", "rpc-client", "rpc-server", "sqlalchemy", "soap", "tornado-client", "tornado-server",
                        "urllib3", "wsgi")

    # Recorder thread for collection/reporting of spans
    thread = None

    def __init__(self):
        self.queue = queue.Queue()

    def start(self):
        """
        This function can be called at first boot or after a fork.  In either case, it will
        assure that the Recorder is in a proper state (via reset()) and spawn a new background
        thread to periodically report queued spans

        Note that this will abandon any previous thread object that (in the case of an `os.fork()`)
        should no longer exist in the forked process.

        (Forked processes carry forward only the thread that called `os.fork()`
        into the new process space.  All other background threads need to be recreated.)

        Calling this directly more than once without an actual fork will cause errors.
        """
        self.reset()
        self.thread.start()

    def reset(self):
        # Prepare the thread for span collection/reporting
        self.thread = threading.Thread(target=self.report_spans)
        self.thread.daemon = True
        self.thread.name = self.THREAD_NAME

    def handle_fork(self):
        self.start()

    def report_spans(self):
        """ Periodically report the queued spans """
        logger.debug(" -> Span reporting thread is now alive")

        def span_work():
            if instana.singletons.agent.should_threads_shutdown.is_set():
                logger.debug("Thread shutdown signal from agent is active: Shutting down span reporting thread")
                return False

            queue_size = self.queue.qsize()
            if queue_size > 0 and instana.singletons.agent.can_send():
                response = instana.singletons.agent.report_traces(self.queued_spans())
                if response:
                    logger.debug("reported %d spans", queue_size)
            return True

        if "INSTANA_TEST" not in os.environ:
            every(2, span_work, "Span Reporting")

    def queue_size(self):
        """ Return the size of the queue; how may spans are queued, """
        return self.queue.qsize()

    def queued_spans(self):
        """ Get all of the spans in the queue """
        span = None
        spans = []
        while True:
            try:
                span = self.queue.get(False)
            except queue.Empty:
                break
            else:
                spans.append(span)
        return spans

    def clear_spans(self):
        """ Clear the queue of spans """
        self.queued_spans()

    def record_span(self, span):
        """
        Convert the passed BasicSpan into and add it to the span queue
        """
        if instana.singletons.agent.can_send() or "INSTANA_TEST" in os.environ:
            source = instana.singletons.agent.get_from_structure()

            if span.operation_name in self.REGISTERED_SPANS:
                json_span = RegisteredSpan(span, source)
            else:
                service_name = instana.singletons.agent.options.service_name
                json_span = SDKSpan(span, source, service_name)

            self.queue.put(json_span)


class AWSLambdaRecorder(StandardRecorder):
    def __init__(self, agent):
        self.agent = agent
        super(AWSLambdaRecorder, self).__init__()

    def record_span(self, span):
        """
        Convert the passed BasicSpan and add it to the span queue
        """
        source = self.agent.get_from_structure()

        if span.operation_name in self.REGISTERED_SPANS:
            json_span = RegisteredSpan(span, source)
        else:
            service_name = self.agent.options.service_name
            json_span = SDKSpan(span, source, service_name)

        # logger.debug("Recorded span: %s", json_span)
        self.agent.collector.span_queue.put(json_span)


class InstanaSampler(Sampler):
    def sampled(self, _):
        return False
